import * as plugins from '../typedserver.plugins.js';

export class Compressor {
  private _cache: Map<string, Buffer>;
  private MAX_CACHE_SIZE: number = 20 * 1024 * 1024; // 20 MB

  constructor() {
    this._cache = new Map<string, Buffer>();
  }

  private _addToCache(key: string, value: Buffer) {
    this._cache.set(key, value);
    this._manageCacheSize();
  }

  private _manageCacheSize() {
    let currentSize = Array.from(this._cache.values()).reduce((acc, buffer) => acc + buffer.length, 0);

    while (currentSize > this.MAX_CACHE_SIZE) {
      const firstKey = this._cache.keys().next().value;
      const firstValue = this._cache.get(firstKey)!;
      currentSize -= firstValue.length;
      this._cache.delete(firstKey);
    }
  }

  public async compressContent(
    content: Buffer,
    method: 'gzip' | 'deflate' | 'br' | 'none'
  ): Promise<Buffer> {
    const cacheKey = content.toString('base64') + method;
    const cachedResult = this._cache.get(cacheKey);

    if (cachedResult) {
      return cachedResult;
    }

    return new Promise((resolve, reject) => {
      const callback = (err: Error | null, result: Buffer) => {
        if (err) reject(err);
        else {
          this._addToCache(cacheKey, result);
          resolve(result);
        }
      };

      switch (method) {
        case 'gzip':
          plugins.zlib.gzip(content, callback);
          break;
        case 'br':
          plugins.zlib.brotliCompress(content, callback);
          break;
        case 'deflate':
          plugins.zlib.deflate(content, callback);
          break;
        default:
          this._addToCache(cacheKey, content);
          resolve(content);
      }
    });
  }

  public determineCompression(acceptEncoding: string | string[]) {
    // Ensure acceptEncoding is a single string
    const encodingString = Array.isArray(acceptEncoding)
      ? acceptEncoding.join(', ')
      : acceptEncoding;

    let compressionMethod: 'gzip' | 'deflate' | 'br' | 'none' = 'none';

    // Check and prioritize compression methods
    if (/\bbr\b/.test(encodingString)) {
      compressionMethod = 'br';
    } else if (/\bgzip\b/.test(encodingString)) {
      compressionMethod = 'gzip';
    } else if (/\bdeflate\b/.test(encodingString)) {
      compressionMethod = 'deflate';
    }

    return compressionMethod;
  }

  public async maybeCompress(requestHeaders: plugins.http.IncomingHttpHeaders, content: Buffer) {
    const acceptEncoding = requestHeaders['accept-encoding'];
    const compressionMethod = this.determineCompression(acceptEncoding);
    const result = await this.compressContent(content, compressionMethod);
    return {
      result,
      compressionMethod,
    };
  }

  public createCompressionStream(method: 'gzip' | 'deflate' | 'br' | 'none') {
    switch (method) {
      case 'gzip':
        return plugins.zlib.createGzip();
      case 'br':
        return plugins.zlib.createBrotliCompress();
      case 'deflate':
        return plugins.zlib.createDeflate();
      default:
        throw new Error('Invalid compression method');
    }
  }
}
